@@ -21,17 +21,18 @@ use std::collections::HashMap;
 use std::time::Duration;
 
 use rtlola_reporting::Span;
+use serde::{Deserialize, Serialize};
 use uom::si::rational64::Frequency as UOM_Frequency;
 
 pub use crate::hir::expression::*;
 pub use crate::modes::ast_conversion::{SpawnDef, TransformationErr};
-pub use crate::modes::dependencies::{DependencyErr, DependencyGraph, EdgeWeight};
+pub use crate::modes::dependencies::{DependencyErr, DependencyGraph, EdgeWeight, Origin};
 pub use crate::modes::memory_bounds::MemorizationBound;
 pub use crate::modes::ordering::{Layer, StreamLayers};
 use crate::modes::HirMode;
 pub use crate::modes::{
     BaseMode, CompleteMode, DepAnaMode, DepAnaTrait, HirStage, MemBoundMode, MemBoundTrait, OrderedMode, OrderedTrait,
-    TypedTrait,
+    TypedMode, TypedTrait,
 };
 use crate::stdlib::FuncDecl;
 pub use crate::type_check::{
@@ -303,13 +304,21 @@ impl<M: HirMode> Hir<M> {
             SRef::Out(o) => {
                 if o < self.outputs.len() {
                     let output = self.outputs.iter().find(|o| o.sr == sr);
-                    output.and_then(|o| o.instance_template.close.map(|e| self.expression(e)))
+                    output.and_then(|o| o.instance_template.close.as_ref().map(|e| self.expression(e.target)))
                 } else {
                     None
                 }
             },
         }
     }
+
+    /// Generates a map from a [StreamReference] to the name of the corresponding stream.
+    pub fn names(&self) -> HashMap<SRef, &str> {
+        self.inputs()
+            .map(|i| (i.sr, i.name.as_str()))
+            .chain(self.outputs().map(|o| (o.sr, o.name.as_str())))
+            .collect()
+    }
 }
 
 /// A collection of maps for expression-related lookups, i.e., expressions, functions, and windows.
@@ -472,7 +481,37 @@ pub(crate) struct InstanceTemplate {
     /// The optional filter condition
     pub(crate) filter: Option<ExprId>,
     /// The optional closing condition
-    pub(crate) close: Option<ExprId>,
+    pub(crate) close: Option<CloseTemplate>,
+}
+
+impl InstanceTemplate {
+    /// Returns a reference to the `Expression` representing the spawn target if it exists
+    pub(crate) fn spawn_target<'a, M: HirMode>(&self, hir: &'a RtLolaHir<M>) -> Option<&'a Expression> {
+        self.spawn
+            .as_ref()
+            .and_then(|st| st.target)
+            .map(|eid| hir.expression(eid))
+    }
+
+    /// Returns a vector of `Expression` references representing the expressions with which the parameters of the stream are initialized
+    pub(crate) fn spawn_arguments<'a, M: HirMode>(&self, hir: &'a RtLolaHir<M>) -> Vec<&'a Expression> {
+        self.spawn_target(hir)
+            .map(|se| {
+                match &se.kind {
+                    ExpressionKind::Tuple(spawns) => spawns.iter().collect(),
+                    _ => vec![se],
+                }
+            })
+            .unwrap_or_else(Vec::new)
+    }
+
+    /// Returns a reference to the `Expression` representing the spawn condition if it exists
+    pub(crate) fn spawn_condition<'a, M: HirMode>(&self, hir: &'a RtLolaHir<M>) -> Option<&'a Expression> {
+        self.spawn
+            .as_ref()
+            .and_then(|st| st.condition)
+            .map(|eid| hir.expression(eid))
+    }
 }
 
 /// Information regarding the spawning behavior of a stream
@@ -486,6 +525,15 @@ pub(crate) struct SpawnTemplate {
     pub(crate) condition: Option<ExprId>,
 }
 
+/// Information regarding the closing behavior of a stream
+#[derive(Debug, Clone)]
+pub(crate) struct CloseTemplate {
+    /// The expression defining if an instance is closed
+    pub(crate) target: ExprId,
+    /// The activation condition describing when an instance is closed
+    pub(crate) pacing: Option<AnnotatedPacingType>,
+}
+
 /// Represents a trigger of an RTLola specification.
 #[derive(Debug, Clone)]
 pub struct Trigger {
@@ -569,7 +617,7 @@ impl AnnotatedType {
 }
 
 /// Allows for referencing a window instance.
-#[derive(Hash, Debug, Clone, Copy, PartialEq, Eq)]
+#[derive(Hash, Debug, Clone, Copy, PartialEq, Eq, Serialize, Deserialize)]
 pub enum WindowReference {
     /// Refers to a sliding window
     Sliding(usize),
@@ -595,7 +643,7 @@ pub type InputReference = usize;
 pub type OutputReference = usize;
 
 /// Allows for referencing a stream within the specification.
-#[derive(Debug, Clone, Copy, Hash, PartialEq, Eq)]
+#[derive(Debug, Clone, Copy, Hash, PartialEq, Eq, Serialize, Deserialize)]
 pub enum StreamReference {
     /// References an input stream.
     In(InputReference),
@@ -650,8 +698,8 @@ impl PartialOrd for StreamReference {
     fn partial_cmp(&self, other: &Self) -> Option<std::cmp::Ordering> {
         use std::cmp::Ordering;
         match (self, other) {
-            (StreamReference::In(i), StreamReference::In(i2)) => Some(i.cmp(&i2)),
-            (StreamReference::Out(o), StreamReference::Out(o2)) => Some(o.cmp(&o2)),
+            (StreamReference::In(i), StreamReference::In(i2)) => Some(i.cmp(i2)),
+            (StreamReference::Out(o), StreamReference::Out(o2)) => Some(o.cmp(o2)),
             (StreamReference::In(_), StreamReference::Out(_)) => Some(Ordering::Less),
             (StreamReference::Out(_), StreamReference::In(_)) => Some(Ordering::Greater),
         }
@@ -662,8 +710,8 @@ impl Ord for StreamReference {
     fn cmp(&self, other: &Self) -> std::cmp::Ordering {
         use std::cmp::Ordering;
         match (self, other) {
-            (StreamReference::In(i), StreamReference::In(i2)) => i.cmp(&i2),
-            (StreamReference::Out(o), StreamReference::Out(o2)) => o.cmp(&o2),
+            (StreamReference::In(i), StreamReference::In(i2)) => i.cmp(i2),
+            (StreamReference::Out(o), StreamReference::Out(o2)) => o.cmp(o2),
             (StreamReference::In(_), StreamReference::Out(_)) => Ordering::Less,
             (StreamReference::Out(_), StreamReference::In(_)) => Ordering::Greater,
         }
@@ -671,7 +719,7 @@ impl Ord for StreamReference {
 }
 
 /// Offset used in the lookup expression
-#[derive(Debug, PartialEq, Eq, Clone, Copy)]
+#[derive(Debug, PartialEq, Eq, Clone, Copy, Hash)]
 pub enum Offset {
     /// A strictly positive discrete offset, e.g., `4`, or `42`
     FutureDiscrete(u32),
@@ -683,6 +731,26 @@ pub enum Offset {
     PastRealTime(Duration),
 }
 
+impl Offset {
+    /// Returns `true`, iff the Offset is negative
+    pub(crate) fn has_negative_offset(&self) -> bool {
+        match self {
+            Offset::FutureDiscrete(_) | Offset::FutureRealTime(_) => false,
+            Offset::PastDiscrete(o) => *o != 0,
+            Offset::PastRealTime(o) => o.as_nanos() != 0,
+        }
+    }
+
+    pub(crate) fn as_memory_bound(&self, dynamic: bool) -> MemorizationBound {
+        match self {
+            Offset::PastDiscrete(o) => MemorizationBound::Bounded(*o) + MemorizationBound::default_value(dynamic),
+            Offset::FutureDiscrete(_) => unimplemented!(),
+            Offset::FutureRealTime(_) => unimplemented!(),
+            Offset::PastRealTime(_) => unimplemented!(),
+        }
+    }
+}
+
 impl PartialOrd for Offset {
     fn partial_cmp(&self, other: &Self) -> Option<std::cmp::Ordering> {
         use std::cmp::Ordering;
