@@ -7,13 +7,15 @@ use std::time::Duration;
 
 use rtlola_parser::ast;
 use rtlola_parser::ast::{FunctionName, Literal as AstLiteral, NodeId, RtLolaAst, SpawnSpec, StreamAccessKind, Type};
-use rtlola_reporting::{Handler, Span};
+use rtlola_reporting::{Diagnostic, RtLolaError, Span};
+use serde::{Deserialize, Serialize};
 
 use super::BaseMode;
 use crate::hir::{
-    AnnotatedPacingType, AnnotatedType, ArithLogOp, Constant as HirConstant, DiscreteAggr, ExprId, Expression,
-    ExpressionKind, ExpressionMaps, FnExprKind, Hir, Inlined, Input, InstanceTemplate, Literal, Offset, Output,
-    Parameter, SRef, SlidingAggr, SpawnTemplate, StreamAccessKind as IRAccess, Trigger, WRef, WidenExprKind, Window,
+    AnnotatedPacingType, AnnotatedType, ArithLogOp, CloseTemplate, Constant as HirConstant, DiscreteAggr, ExprId,
+    Expression, ExpressionKind, ExpressionMaps, FnExprKind, Hir, Inlined, Input, InstanceTemplate, Literal, Offset,
+    Output, Parameter, SRef, SlidingAggr, SpawnTemplate, StreamAccessKind as IRAccess, Trigger, WRef, WidenExprKind,
+    Window,
 };
 use crate::modes::ast_conversion::naming::{Declaration, NamingAnalysis};
 use crate::stdlib::FuncDecl;
@@ -27,9 +29,9 @@ impl Hir<BaseMode> {
     /// - Checks for proper expression kinds within all expressions.
     /// - Ensures no missing expressions and inlines all constant definitions, see [Constant](crate::hir:Constant).
     /// - Assigns new expression ids to all expressions.
-    pub(crate) fn from_ast(ast: RtLolaAst, handler: &Handler) -> Result<Self, TransformationErr> {
-        let mut naming_analyzer = NamingAnalysis::new(&handler);
-        let decl_table = naming_analyzer.check(&ast);
+    pub(crate) fn from_ast(ast: RtLolaAst) -> Result<Self, RtLolaError> {
+        let mut naming_analyzer = NamingAnalysis::new();
+        let decl_table = naming_analyzer.check(&ast)?;
         let func_table: HashMap<String, FuncDecl> = decl_table
             .values()
             .filter(|decl| matches!(decl, Declaration::Func(_)))
@@ -53,7 +55,7 @@ impl Hir<BaseMode> {
             stream_by_name.insert(i.name.name.clone(), sr);
         }
         let stream_by_name = stream_by_name;
-        ExpressionTransformer::run(decl_table, stream_by_name, ast, func_table)
+        ExpressionTransformer::run(decl_table, stream_by_name, ast, func_table).map_err(|e| e.into_diagnostic().into())
     }
 }
 
@@ -62,18 +64,18 @@ impl Hir<BaseMode> {
 pub type SpawnDef<'a> = (Option<&'a Expression>, Option<&'a Expression>);
 
 /// A [TransformationErr] describes the kind off error raised during the Ast to Hir conversion.
-#[derive(Debug, Clone)]
+#[derive(Debug, Clone, Serialize, Deserialize)]
 pub enum TransformationErr {
     /// A function was found when a stream was expected.
-    InvalidIdentRef(FunctionName),
+    InvalidIdentRef(Span, FunctionName),
     /// No valid streamrefernce found while it was expected.
-    InvalidRefExpr(String),
+    InvalidRefExpr(Span, String),
     /// A declared constant had no type annotation.
     ConstantWithoutType(Span),
     /// Could not parse numeric literal.
     NonNumericInLiteral(Span),
     /// Invalid activation condition.
-    InvalidAc(String),
+    InvalidAc(Span, String),
     /// Offset expression of realtime offset could not be parsed as frequency.
     InvalidRealtimeOffset(Span),
     /// Window duration could not be parsed into correct type.
@@ -83,15 +85,116 @@ pub enum TransformationErr {
     /// Widen call expects a single type argument.
     MissingWidenArg(Span),
     /// Annotated type could not be matched.
-    InvalidTypeArgument(Type, Span),
-    /// Called functino unknown.
+    InvalidType(Type, String, Span),
+    /// Called function is unknown.
     UnknownFunction(Span),
-    /// Input stream had no type annotation.
-    MissingInputType(Span),
-    /// Object method called, currently unimplemented.
-    MethodAccess,
     /// Non duration literal with postfix found.
     InvalidLiteral(Span),
+    /// An access to a parameterized output stream is missing its parameters
+    MissingArguments(Span),
+    /// An output stream with parameters is missing a spawn target
+    MissingSpawn(Span),
+    /// An output stream with a spawn target is missing parameters
+    MissingParameters(Span),
+    /// The number of parameters of an output differs from the number of spawn expressions.
+    /// Span of the output, number of parameters, number of spawn expressions
+    SpawnParameterMismatch(Span, usize, usize),
+}
+
+impl TransformationErr {
+    pub(crate) fn into_diagnostic(self) -> Diagnostic {
+        match self {
+            TransformationErr::InvalidIdentRef(span, name) => {
+                Diagnostic::error("Expected a stream reference, but found a function.").add_span_with_label(
+                    span,
+                    Some(&format!("Found function {} here", name)),
+                    true,
+                )
+            },
+            TransformationErr::InvalidRefExpr(span, reason) => {
+                Diagnostic::error(&format!("Invalid stream identifier: {}", reason)).add_span_with_label(
+                    span,
+                    Some("Found here"),
+                    true,
+                )
+            },
+            TransformationErr::ConstantWithoutType(span) => {
+                Diagnostic::error("Missing type annotation of constant stream.").add_span_with_label(
+                    span,
+                    Some("here"),
+                    true,
+                )
+            },
+            TransformationErr::NonNumericInLiteral(span) => {
+                Diagnostic::error("Invalid numeric literal.").add_span_with_label(span, Some("here"), true)
+            },
+            TransformationErr::InvalidAc(span, reason) => {
+                Diagnostic::error(&format!("Invalid frequency annotation: {}", reason)).add_span_with_label(
+                    span,
+                    Some("here"),
+                    true,
+                )
+            },
+            TransformationErr::InvalidRealtimeOffset(span) => {
+                Diagnostic::error("Invalid time format.").add_span_with_label(span, Some("here"), true)
+            },
+            TransformationErr::InvalidDuration(reason, span) => {
+                Diagnostic::error("Invalid window duration.").add_span_with_label(span, Some(&reason), true)
+            },
+            TransformationErr::MissingExpr(span) => {
+                Diagnostic::error("Expected an expression.").add_span_with_label(span, Some("here"), true)
+            },
+            TransformationErr::MissingWidenArg(span) => {
+                Diagnostic::error("The widen expression expects an argument.").add_span_with_label(
+                    span,
+                    Some("missing argument here"),
+                    true,
+                )
+            },
+            TransformationErr::InvalidType(ty, reason, span) => {
+                Diagnostic::error(&format!("Unknown type {}.", ty)).add_span_with_label(span, Some(&reason), true)
+            },
+            TransformationErr::UnknownFunction(span) => {
+                Diagnostic::error("Unknown function.").add_span_with_label(span, Some("Found here"), true)
+            },
+            TransformationErr::InvalidLiteral(span) => {
+                Diagnostic::error("Unit declarations are not allowed on non-time literals.").add_span_with_label(
+                    span,
+                    Some("here"),
+                    true,
+                )
+            },
+            TransformationErr::MissingArguments(span) => {
+                Diagnostic::error("An access to a parameterized output stream is missing its arguments.").add_span_with_label(
+                    span,
+                    Some("here"),
+                    true,
+                )
+            }
+            TransformationErr::MissingSpawn(span) => {
+                Diagnostic::error("The following stream has parameters but no spawn target.").add_span_with_label(
+                    span,
+                    Some("here"),
+                    true,
+                )
+                    .add_note("Help: Add a spawn declaration of the form: spawn with (exp1, ..., exp_n) if ...")
+            }
+            TransformationErr::MissingParameters(span) => {
+                Diagnostic::error("The following stream has a spawn declaration but no parameters.").add_span_with_label(
+                    span,
+                    Some("here"),
+                    true,
+                )
+            }
+            TransformationErr::SpawnParameterMismatch(span, paras, targets) => {
+                Diagnostic::error(&format!("The number of parameters of the stream differs from the number of spawn target expressions. Found {} parameters and {} spawn targets", paras, targets)).add_span_with_label(
+                    span,
+                    Some("here"),
+                    true,
+                )
+            }
+        }
+    }
 }
 
 #[derive(Debug)]
@@ -148,26 +251,66 @@ impl ExpressionTransformer {
                 annotated_pacing_type,
                 ..
             } = (*o).clone();
-            let params: Vec<Parameter> = params
+            let params = params
                 .iter()
                 .enumerate()
                 .map(|(ix, p)| {
                     assert_eq!(ix, p.param_idx);
-                    Parameter {
-                        name: p.name.name.clone(),
-                        annotated_type: p.ty.as_ref().and_then(Self::annotated_type),
-                        idx: p.param_idx,
-                        span: p.span.clone(),
-                    }
+                    p.ty.as_ref()
+                        .map_or(Ok(None), |ty| {
+                            Self::annotated_type(ty)
+                                .map(Some)
+                                .map_err(|reason| (reason, ty.clone(), p.span.clone()))
+                        })
+                        .map(|p_ty| {
+                            Parameter {
+                                name: p.name.name.clone(),
+                                annotated_type: p_ty,
+                                idx: p.param_idx,
+                                span: p.span.clone(),
+                            }
+                        })
                 })
-                .collect();
-            let annotated_type = annotated_type.as_ref().and_then(Self::annotated_type);
+                .collect::<Result<Vec<Parameter>, (String, Type, Span)>>()
+                .map_err(|(reason, ty, span)| TransformationErr::InvalidType(ty, reason, span))?;
+            let annotated_type = annotated_type
+                .as_ref()
+                .map_or(Ok(None), |ty| {
+                    Self::annotated_type(ty)
+                        .map(Some)
+                        .map_err(|reason| (reason, ty.clone(), ty.span.clone()))
+                })
+                .map_err(|(reason, ty, span)| TransformationErr::InvalidType(ty, reason, span))?;
             let expression = self.transform_expression(expression, sr)?;
             let expr_id = expression.eid;
             exprid_to_expr.insert(expr_id, expression);
             let annotated_pacing_type = annotated_pacing_type
                 .map_or(Ok(None), |pt| self.transform_pt(&mut exprid_to_expr, pt, sr).map(Some))?;
             let instance_template = self.transform_template_spec(spawn, filter, close, &mut exprid_to_expr, sr)?;
+
+            //Check that if the output has parameters it has a spawn condition with a target and the other way around
+            if !params.is_empty() && instance_template.spawn.as_ref().and_then(|s| s.target).is_none() {
+                return Err(TransformationErr::MissingSpawn(o.span.clone()));
+            }
+            if let Some(target) = instance_template.spawn.as_ref().and_then(|s| s.target) {
+                let spawn_expr = &exprid_to_expr[&target];
+                if params.is_empty() {
+                    return Err(TransformationErr::MissingParameters(spawn_expr.span.clone()));
+                }
+                // check that they are equal length
+                let num_spawn_expr = match &spawn_expr.kind {
+                    ExpressionKind::Tuple(elements) => elements.len(),
+                    _ => 1,
+                };
+                if num_spawn_expr != params.len() {
+                    return Err(TransformationErr::SpawnParameterMismatch(
+                        o.span.clone(),
+                        params.len(),
+                        num_spawn_expr,
+                    ));
+                }
+            }
+
             hir_outputs.push(Output {
                 name: name.name,
                 sr,
@@ -213,7 +356,7 @@ impl ExpressionTransformer {
             .map(|(ix, i)| {
                 Ok(Input {
                     annotated_type: Self::annotated_type(&i.ty)
-                        .ok_or_else(|| TransformationErr::MissingInputType(i.span.clone()))?,
+                        .map_err(|reason| TransformationErr::InvalidType(i.ty.clone(), reason, i.span.clone()))?,
                     name: i.name.name.clone(),
                     sr: SRef::In(ix),
                     span: i.span.clone(),
@@ -243,58 +386,55 @@ impl ExpressionTransformer {
         })
     }
 
-    fn annotated_type(ast_ty: &Type) -> Option<AnnotatedType> {
+    fn annotated_type(ast_ty: &Type) -> Result<AnnotatedType, String> {
         use rtlola_parser::ast::TypeKind;
         match &ast_ty.kind {
             TypeKind::Tuple(vec) => {
-                Some(AnnotatedType::Tuple(
-                    vec.iter()
-                        .map(|inner| Self::annotated_type(inner).expect("Inner types can not be missing"))
-                        .collect(),
-                ))
-            },
-            TypeKind::Optional(inner) => {
-                Some(AnnotatedType::Option(
-                    Self::annotated_type(inner)
-                        .expect("Inner types can not be missing")
-                        .into(),
-                ))
+                let inner: Result<Vec<AnnotatedType>, String> = vec.iter().map(Self::annotated_type).collect();
+                inner.map(AnnotatedType::Tuple)
             },
+            TypeKind::Optional(inner) => Self::annotated_type(inner).map(|inner| AnnotatedType::Option(inner.into())),
             TypeKind::Simple(string) => {
                 if string == "String" {
-                    return Some(AnnotatedType::String);
+                    return Ok(AnnotatedType::String);
                 }
                 if string == "Bool" {
-                    return Some(AnnotatedType::Bool);
+                    return Ok(AnnotatedType::Bool);
                 }
                 if let Some(size_str) = string.strip_prefix("Int") {
                     if string.len() == 3 {
-                        return Some(AnnotatedType::Int(64));
+                        return Ok(AnnotatedType::Int(64));
                     } else {
-                        let size: u32 = size_str.parse().expect("Invalid char followed Int type annotation");
-                        return Some(AnnotatedType::Int(size));
+                        let size: u32 = size_str
+                            .parse()
+                            .map_err(|_| "Invalid char followed Int type annotation".to_string())?;
+                        return Ok(AnnotatedType::Int(size));
                     }
                 }
                 if let Some(size_str) = string.strip_prefix("UInt") {
                     if string.len() == 4 {
-                        return Some(AnnotatedType::UInt(64));
+                        return Ok(AnnotatedType::UInt(64));
                     } else {
-                        let size: u32 = size_str.parse().expect("Invalid char followed UInt type annotation");
-                        return Some(AnnotatedType::UInt(size));
+                        let size: u32 = size_str
+                            .parse()
+                            .map_err(|_| "Invalid char followed UInt type annotation".to_string())?;
+                        return Ok(AnnotatedType::UInt(size));
                     }
                 }
                 if let Some(size_str) = string.strip_prefix("Float") {
                     if string.len() == 5 {
-                        return Some(AnnotatedType::Float(64));
+                        return Ok(AnnotatedType::Float(64));
                     } else {
-                        let size: u32 = size_str.parse().expect("Invalid char followed Float type annotation");
-                        return Some(AnnotatedType::Float(size));
+                        let size: u32 = size_str
+                            .parse()
+                            .map_err(|_| "Invalid char followed Float type annotation".to_string())?;
+                        return Ok(AnnotatedType::Float(size));
                     }
                 }
                 if string == "Bytes" {
-                    return Some(AnnotatedType::Bytes);
+                    return Ok(AnnotatedType::Bytes);
                 }
-                None
+                Err("unknown type".into())
             },
         }
     }
@@ -309,10 +449,12 @@ impl ExpressionTransformer {
                 match &self.decl_table[&expr.id] {
                     Declaration::In(i) => Ok((self.stream_by_name[&i.name.name], Vec::new())),
                     Declaration::Out(o) => Ok((self.stream_by_name[&o.name.name], Vec::new())),
+                    Declaration::ParamOut(_) => Err(TransformationErr::MissingArguments(expr.span.clone())),
                     _ => {
-                        Err(TransformationErr::InvalidRefExpr(String::from(
-                            "Non-identifier transformed to SRef",
-                        )))
+                        Err(TransformationErr::InvalidRefExpr(
+                            expr.span.clone(),
+                            String::from("Non-identifier transformed to SRef"),
+                        ))
                     },
                 }
             },
@@ -326,10 +468,15 @@ impl ExpressionTransformer {
                                 .collect::<Result<Vec<_>, TransformationErr>>()?,
                         ))
                     },
-                    _ => Err(TransformationErr::InvalidIdentRef(name.clone())),
+                    _ => Err(TransformationErr::InvalidIdentRef(expr.span.clone(), name.clone())),
                 }
             },
-            _ => Err(TransformationErr::InvalidRefExpr(format!("{:?}", expr.kind))),
+            _ => {
+                Err(TransformationErr::InvalidRefExpr(
+                    expr.span.clone(),
+                    format!("{:?}", expr.kind),
+                ))
+            },
         }
     }
 
@@ -381,7 +528,9 @@ impl ExpressionTransformer {
     ) -> Result<AnnotatedPacingType, TransformationErr> {
         if let ast::ExpressionKind::Lit(l) = &pt_expr.kind {
             if let ast::LitKind::Numeric(_, Some(_)) = &l.kind {
-                let val = pt_expr.parse_freqspec().map_err(TransformationErr::InvalidAc)?;
+                let val = pt_expr
+                    .parse_freqspec()
+                    .map_err(|reason| TransformationErr::InvalidAc(pt_expr.span.clone(), reason))?;
                 return Ok(AnnotatedPacingType::Frequency {
                     span: pt_expr.span.clone(),
                     value: val,
@@ -421,7 +570,7 @@ impl ExpressionTransformer {
                             c.ty.as_ref()
                                 .ok_or_else(|| TransformationErr::ConstantWithoutType(span.clone()))?;
                         let annotated_type = Self::annotated_type(ty)
-                            .ok_or_else(|| TransformationErr::ConstantWithoutType(span.clone()))?;
+                            .map_err(|reason| TransformationErr::InvalidType(ty.clone(), reason, span.clone()))?;
                         ExpressionKind::LoadConstant(HirConstant::Inlined(Inlined {
                             lit: self.transform_literal(&c.literal)?,
                             ty: annotated_type,
@@ -429,7 +578,10 @@ impl ExpressionTransformer {
                     },
 
                     Declaration::Param(p) => ExpressionKind::ParameterAccess(current_output, p.param_idx),
-                    Declaration::ParamOut(_) | Declaration::Func(_) | Declaration::Type(_) => {
+                    Declaration::ParamOut(_) => {
+                        return Err(TransformationErr::MissingArguments(span));
+                    },
+                    Declaration::Func(_) | Declaration::Type(_) => {
                         unreachable!("Identifiers can only refer to streams")
                     },
                 }
@@ -486,8 +638,8 @@ impl ExpressionTransformer {
                 wait,
                 aggregation: win_op,
             } => {
-                let (sref, _) = self.get_stream_ref(&w_expr, current_output)?;
-                let idx = self.sliding_windows.len();
+                let (sref, paras) = self.get_stream_ref(&w_expr, current_output)?;
+                let idx = self.discrete_windows.len();
                 let wref = WRef::Discrete(idx);
                 let duration = (*duration)
                     .parse_discrete_duration()
@@ -504,7 +656,7 @@ impl ExpressionTransformer {
                     eid: new_id,
                 };
                 self.discrete_windows.push(window);
-                ExpressionKind::StreamAccess(sref, IRAccess::DiscreteWindow(WRef::Discrete(idx)), Vec::new())
+                ExpressionKind::StreamAccess(sref, IRAccess::DiscreteWindow(WRef::Discrete(idx)), paras)
             },
             ast::ExpressionKind::SlidingWindowAggregation {
                 expr: w_expr,
@@ -512,7 +664,7 @@ impl ExpressionTransformer {
                 wait,
                 aggregation: win_op,
             } => {
-                let (sref, _) = self.get_stream_ref(&w_expr, current_output)?;
+                let (sref, paras) = self.get_stream_ref(&w_expr, current_output)?;
                 let idx = self.sliding_windows.len();
                 let wref = WRef::Sliding(idx);
                 let duration = Self::parse_duration_from_expr(&*duration)
@@ -529,7 +681,7 @@ impl ExpressionTransformer {
                     eid: new_id,
                 };
                 self.sliding_windows.push(window);
-                ExpressionKind::StreamAccess(sref, IRAccess::SlidingWindow(WRef::Sliding(idx)), Vec::new())
+                ExpressionKind::StreamAccess(sref, IRAccess::SlidingWindow(WRef::Sliding(idx)), paras)
             },
             ast::ExpressionKind::Binary(op, left, right) => {
                 use rtlola_parser::ast::BinOp;
@@ -658,7 +810,7 @@ impl ExpressionTransformer {
                         ty: match type_param.get(0) {
                             Some(t) => {
                                 Self::annotated_type(t)
-                                    .ok_or_else(|| TransformationErr::InvalidTypeArgument(t.clone(), span.clone()))?
+                                    .map_err(|reason| TransformationErr::InvalidType(t.clone(), reason, span.clone()))?
                             },
                             None => todo!("error case"),
                         },
@@ -671,7 +823,7 @@ impl ExpressionTransformer {
                             .into_iter()
                             .map(|t| {
                                 Self::annotated_type(&t)
-                                    .ok_or_else(|| TransformationErr::InvalidTypeArgument(t, span.clone()))
+                                    .map_err(|reason| TransformationErr::InvalidType(t, reason, span.clone()))
                             })
                             .collect::<Result<Vec<_>, TransformationErr>>()?,
                     }))
@@ -774,10 +926,14 @@ impl ExpressionTransformer {
                 )))
             })?,
             close: close_spec.map_or(Ok(None), |close_spec| {
-                Ok(Some(Self::insert_return(
+                let pacing = close_spec.annotated_pacing.map_or(Ok(None), |pt| {
+                    Ok(Some(self.transform_pt(exprid_to_expr, pt, current_output)?))
+                })?;
+                let target = Self::insert_return(
                     exprid_to_expr,
                     self.transform_expression(close_spec.target, current_output)?,
-                )))
+                );
+                Ok(Some(CloseTemplate { target, pacing }))
             })?,
         })
     }
@@ -785,19 +941,17 @@ impl ExpressionTransformer {
 
 #[cfg(test)]
 mod tests {
-    use std::path::PathBuf;
+    use std::collections::HashSet;
 
     use rtlola_parser::ast::WindowOperation;
-    use rtlola_parser::{parse_with_handler, ParserConfig};
+    use rtlola_parser::{parse, ParserConfig};
 
     use super::*;
-    use crate::hir::StreamAccessKind;
+    use crate::hir::{ExpressionContext, StreamAccessKind};
 
     fn obtain_expressions(spec: &str) -> Hir<BaseMode> {
-        let handler = Handler::new(PathBuf::new(), spec.into());
-        let ast = parse_with_handler(ParserConfig::for_string(spec.to_string()), &handler)
-            .unwrap_or_else(|e| panic!("{}", e));
-        crate::from_ast(ast, &handler).unwrap()
+        let ast = parse(ParserConfig::for_string(spec.to_string())).unwrap_or_else(|e| panic!("{:?}", e));
+        crate::from_ast(ast).unwrap()
     }
 
     #[test]
@@ -812,10 +966,10 @@ mod tests {
         output o4 := 3 + 4
         output o5 := o
         output f := sqrt(4)
-        output p (x,y) := x
+        output p (x,y) spawn with (i,i) if true := x
         output t := (1,3)
         output t1 := t.0
-        output off := o.defaults(to:2)
+        output def := o.defaults(to:2)
         output off := o.offset(by:-1)
         output w := o2.aggregate(over:3s , using: sum)";
         let _ir = obtain_expressions(spec);
@@ -902,7 +1056,7 @@ mod tests {
 
     #[test]
     fn parameter_expr() {
-        let spec = "output o(a,b,c) :=  if c then a else b";
+        let spec = "output o(a,b,c) spawn @1Hz with (1, 2, 3) :=  if c then a else b";
         let ir = obtain_expressions(spec);
         let output_expr_id = ir.outputs[0].expr_id;
         let expr = &ir.expression(output_expr_id);
@@ -924,7 +1078,8 @@ mod tests {
     #[test]
     fn parametrized_access() {
         use crate::hir::StreamAccessKind;
-        let spec = "output o(a,b,c) :=  if c then a else b output A := o(1,2,true).offset(by:-1)";
+        let spec =
+            "output o(a,b,c) spawn @1Hz with (1, 2, true):=  if c then a else b output A := o(1,2,true).offset(by:-1)";
         let ir = obtain_expressions(spec);
         let output_expr_id = ir.outputs[1].expr_id;
         let expr = &ir.expression(output_expr_id);
@@ -1031,7 +1186,7 @@ mod tests {
     #[test]
     fn functions() {
         use crate::hir::StreamAccessKind;
-        let spec = "import math output o(a: Int) := max(3,4) output c := o(1)";
+        let spec = "import math output o(a: Int) spawn @1Hz with 3:= max(3,4) output c := o(1)";
         let ir = obtain_expressions(spec);
         //check that this functions exists
         let _ = ir.func_declaration("max");
@@ -1045,7 +1200,7 @@ mod tests {
 
     #[test]
     fn function_param_default() {
-        let spec = "import math output o(a: Int) := sqrt(a) output c := o(1).defaults(to:1)";
+        let spec = "import math output o(a: Int) spawn @1Hz with 3 := sqrt(a) output c := o(1).defaults(to:1)";
         let ir = obtain_expressions(spec);
         //check purely for valid access
         let _ = ir.func_declaration("sqrt");
@@ -1164,4 +1319,89 @@ mod tests {
             Some(AnnotatedPacingType::Frequency { .. })
         ))
     }
+
+    #[test]
+    fn test_instance_window() {
+        let spec = "input a: Int32\n\
+        output b (p: Bool) spawn with a = 42 := a\n\
+        output c @ 1Hz := b(false).aggregate(over: 1s, using: Σ)\n";
+        let ir = obtain_expressions(spec);
+        let expr = &ir.expression(ir.outputs[1].expr_id).kind;
+        assert!(
+            matches!(expr, ExpressionKind::StreamAccess(_, StreamAccessKind::SlidingWindow(_), paras) if paras.len() == 1)
+        );
+    }
+
+    #[test]
+    #[should_panic]
+    fn test_missing_spawn() {
+        let spec = "input a: Int32\n\
+        output b (p: Bool) := a";
+        obtain_expressions(spec);
+    }
+
+    #[test]
+    #[should_panic]
+    fn test_missing_parameters() {
+        let spec = "input a: Int32\n\
+        output b spawn with a := a";
+        obtain_expressions(spec);
+    }
+
+    #[test]
+    #[should_panic]
+    fn test_parameter_spawn_mismatch() {
+        let spec = "input a: Int32\n\
+        output b (p1, p2) spawn with a := a";
+        obtain_expressions(spec);
+    }
+
+    #[test]
+    fn test_expression_context() {
+        macro_rules! para_map {
+            ($($x:expr),+ $(,)?) => (
+                vec![$($x),+].into_iter().map(|(k, set)| (k, set.into_iter().collect::<HashSet<usize>>())).collect::<HashMap<(SRef, usize), HashSet<usize>>>()
+            );
+        }
+
+        let spec = "input a: Int32\n\
+        output b (p1, p2, p3) spawn with (a, a.hold(or: 5), a.offset(by: -5).defaults(to: 7)) if a = 5 := a + p1 + p2 + p3\n\
+        output c (p1, p2, p3, p4) spawn with (a.hold(or: 5), a, a.hold(or: 5), a.offset(by: -5).defaults(to: 8)) if a = 5 := a + p1 + p2 + p3\n\
+        output d (p1, p2) spawn with (a.hold(or: 5), a) := a + p1 + p2\n";
+        let ir = obtain_expressions(spec);
+        let ctx = ExpressionContext::new(&ir);
+
+        let b = SRef::Out(0);
+        let c = SRef::Out(1);
+        let d = SRef::Out(2);
+
+        let b_map = ctx.map_for(b).clone();
+        assert_eq!(
+            b_map,
+            para_map![
+                ((b, 0), vec![0]),
+                ((b, 1), vec![1]),
+                ((b, 2), vec![2]),
+                ((c, 0), vec![1]),
+                ((c, 1), vec![0]),
+                ((c, 2), vec![1]),
+            ]
+        );
+
+        let c_map = ctx.map_for(c).clone();
+        assert_eq!(
+            c_map,
+            para_map![
+                ((c, 0), vec![0, 2]),
+                ((c, 1), vec![1]),
+                ((c, 2), vec![0, 2]),
+                ((c, 3), vec![3]),
+                ((b, 0), vec![1]),
+                ((b, 1), vec![0, 2]),
+            ]
+        );
+
+        let d_map = ctx.map_for(d).clone();
+        assert_eq!(d_map, para_map![((d, 0), vec![0]), ((d, 1), vec![1])]);
+    }
 }
