@@ -4,7 +4,7 @@ use std::collections::HashMap;
 use std::rc::Rc;
 
 use rtlola_parser::ast::{Ident, NodeId, *};
-use rtlola_reporting::{Diagnostic, Handler, Span};
+use rtlola_reporting::{Diagnostic, RtLolaError, Span};
 
 use crate::hir::AnnotatedType;
 use crate::stdlib::FuncDecl;
@@ -28,17 +28,16 @@ pub(crate) type DeclarationTable = HashMap<NodeId, Declaration>;
 /// - used identifiers have valid definition
 /// - all type annotations describe a valid type
 #[derive(Debug)]
-pub struct NamingAnalysis<'b> {
+pub struct NamingAnalysis {
     declarations: ScopedDecl,
     type_declarations: ScopedDecl,
     fun_declarations: ScopedDecl,
     result: DeclarationTable,
-    handler: &'b Handler,
 }
 
-impl<'b> NamingAnalysis<'b> {
+impl NamingAnalysis {
     /// Constructs a new [NamingAnalysis] to perform checks on all identifiers. Use [check](NamingAnalysis::check) to start.
-    pub fn new(handler: &'b Handler) -> Self {
+    pub fn new() -> Self {
         let mut scoped_decls = ScopedDecl::new();
 
         for (name, ty) in AnnotatedType::primitive_types() {
@@ -53,7 +52,6 @@ impl<'b> NamingAnalysis<'b> {
             type_declarations: scoped_decls,
             fun_declarations: ScopedDecl::new(),
             result: HashMap::new(),
-            handler,
         }
     }
 
@@ -62,8 +60,9 @@ impl<'b> NamingAnalysis<'b> {
     /// Checks if
     /// * name of declaration is a keyword
     /// * declaration already exists in current scope
-    fn add_decl_for(&mut self, decl: Declaration) {
+    fn add_decl_for(&mut self, decl: Declaration) -> Result<(), RtLolaError> {
         assert!(!decl.is_type());
+        let mut error = RtLolaError::new();
         let name = decl
             .get_name()
             .expect("added declarations are guaranteed to have a name");
@@ -73,11 +72,13 @@ impl<'b> NamingAnalysis<'b> {
         // check for keyword
         let lower = name.to_lowercase();
         if KEYWORDS.contains(&lower.as_str()) {
-            self.handler.error_with_span(
-                &format!("`{}` is a reserved keyword", name),
-                span.clone(),
-                Some("use a different name here"),
-            )
+            error.add(
+                Diagnostic::error(&format!("`{}` is a reserved keyword", name)).add_span_with_label(
+                    span.clone(),
+                    Some("use a different name here"),
+                    true,
+                ),
+            );
         }
 
         let decl_name = if let Declaration::Func(rcfunc) = &decl {
@@ -86,21 +87,25 @@ impl<'b> NamingAnalysis<'b> {
             DeclName::Ident(name.to_string())
         };
         if let Some(decl) = self.declarations.get_decl_in_current_scope_for(&decl_name) {
-            Diagnostic::error(self.handler, &format!("the name `{}` is defined multiple times", name))
-                .add_span_with_label(span, Some(&format!("`{}` redefined here", name)), true)
-                .maybe_add_span_with_label(
-                    decl.get_span(),
-                    Some(&format!("previous definition of the value `{}` here", name)),
-                    false,
-                )
-                .emit();
+            error.add(
+                Diagnostic::error(&format!("the name `{}` is defined multiple times", name))
+                    .add_span_with_label(span, Some(&format!("`{}` redefined here", name)), true)
+                    .maybe_add_span_with_label(
+                        decl.get_span(),
+                        Some(&format!("previous definition of the value `{}` here", name)),
+                        false,
+                    ),
+            );
         } else {
             self.declarations.add_decl_for(name, decl.clone());
         }
+        Result::from(error)?;
+        Ok(())
     }
 
     /// Checks if given type is bound
-    fn check_type(&mut self, ty: &Type) {
+    fn check_type(&mut self, ty: &Type) -> Result<(), RtLolaError> {
+        let mut error = RtLolaError::new();
         match &ty.kind {
             TypeKind::Simple(name) => {
                 if let Some(decl) = self.type_declarations.get_decl_for(&DeclName::Ident(name.to_string())) {
@@ -108,24 +113,35 @@ impl<'b> NamingAnalysis<'b> {
                     self.result.insert(ty.id, decl);
                 } else {
                     // it does not exist
-                    self.handler.error_with_span(
-                        &format!("cannot find type `{}` in this scope", name),
-                        ty.span.clone(),
-                        Some("not found in this scope"),
+                    error.add(
+                        Diagnostic::error(&format!("cannot find type `{}` in this scope", name)).add_span_with_label(
+                            ty.span.clone(),
+                            Some("not found in this scope"),
+                            true,
+                        ),
                     );
                 }
             },
             TypeKind::Tuple(elements) => {
                 elements.iter().for_each(|ty| {
-                    self.check_type(ty);
+                    if let Err(e) = self.check_type(ty) {
+                        error.join(e);
+                    }
                 })
             },
-            TypeKind::Optional(ty) => self.check_type(ty),
+            TypeKind::Optional(ty) => {
+                if let Err(e) = self.check_type(ty) {
+                    error.join(e);
+                }
+            },
         }
+        Result::from(error)?;
+        Ok(())
     }
 
     /// Checks that the parameter name and type are both valid
-    fn check_param(&mut self, param: &Rc<Parameter>) {
+    fn check_param(&mut self, param: &Rc<Parameter>) -> Result<(), RtLolaError> {
+        let mut error = RtLolaError::new();
         // check the name
         if let Some(decl) = self
             .declarations
@@ -138,92 +154,123 @@ impl<'b> NamingAnalysis<'b> {
                 .declarations
                 .get_decl_in_current_scope_for(&DeclName::Ident(param.name.name.clone()))
             {
-                Diagnostic::error(
-                    self.handler,
-                    &format!(
+                error.add(
+                    Diagnostic::error(&format!(
                         "identifier `{}` is use more than once in this paramater list",
                         param.name.name
+                    ))
+                    .add_span_with_label(
+                        param.name.span.clone(),
+                        Some(&format!("`{}` used as a parameter more than once", param.name.name)),
+                        true,
+                    )
+                    .add_span_with_label(
+                        decl.get_span().expect("as it is in parameter list, it has a span"),
+                        Some(&format!("previous use of the parameter `{}` here", param.name.name)),
+                        false,
                     ),
-                )
-                .add_span_with_label(
-                    param.name.span.clone(),
-                    Some(&format!("`{}` used as a parameter more than once", param.name.name)),
-                    true,
-                )
-                .add_span_with_label(
-                    decl.get_span().expect("as it is in parameter list, it has a span"),
-                    Some(&format!("previous use of the parameter `{}` here", param.name.name)),
-                    false,
-                )
-                .emit();
+                );
             }
         } else {
             // it does not exist
-            self.add_decl_for(Declaration::Param(param.clone()));
+            if let Err(e) = self.add_decl_for(Declaration::Param(param.clone())) {
+                error.join(e);
+            }
         }
 
         // check the type is there exists a parameter type
         if let Some(param_ty) = param.ty.as_ref() {
-            self.check_type(param_ty)
+            if let Err(e) = self.check_type(param_ty) {
+                error.join(e);
+            }
         }
+
+        Result::from(error)
     }
 
     /// Entry method, checks that every identifier in the given spec is bound.
-    pub(crate) fn check(&mut self, spec: &RtLolaAst) -> DeclarationTable {
+    pub(crate) fn check(&mut self, spec: &RtLolaAst) -> Result<DeclarationTable, RtLolaError> {
         use crate::stdlib;
+        let mut error = RtLolaError::new();
         self.fun_declarations.add_all_fun_decl(stdlib::implicit_module());
         for import in &spec.imports {
             match import.name.name.as_str() {
                 "math" => self.fun_declarations.add_all_fun_decl(stdlib::math_module()),
                 "regex" => self.fun_declarations.add_all_fun_decl(stdlib::regex_module()),
                 n => {
-                    self.handler.error_with_span(
-                        &format!("unresolved import `{}`", n),
-                        import.name.span.clone(),
-                        Some(&format!("no `{}` in the root", n)),
-                    )
+                    error.add(
+                        Diagnostic::error(&format!("unresolved import `{}`", n)).add_span_with_label(
+                            import.name.span.clone(),
+                            Some(&format!("no `{}` in the root", n)),
+                            true,
+                        ),
+                    );
                 },
             }
         }
 
         // Store global declarations, i.e., constants, inputs, and outputs of the given specification
         for constant in &spec.constants {
-            self.add_decl_for(Declaration::Const(constant.clone()));
+            if let Err(e) = self.add_decl_for(Declaration::Const(constant.clone())) {
+                error.join(e);
+            }
             if let Some(ty) = constant.ty.as_ref() {
-                self.check_type(ty)
+                if let Err(e) = self.check_type(ty) {
+                    error.join(e);
+                }
             }
         }
 
         for input in &spec.inputs {
-            self.add_decl_for(Declaration::In(input.clone()));
-            self.check_type(&input.ty);
+            if let Err(e) = self.add_decl_for(Declaration::In(input.clone())) {
+                error.join(e);
+            }
+            if let Err(e) = self.check_type(&input.ty) {
+                error.join(e);
+            }
 
             // check types for parametric inputs
             self.declarations.push();
-            input.params.iter().for_each(|param| self.check_param(param));
+            let err = input
+                .params
+                .iter()
+                .flat_map(|param| self.check_param(param).err())
+                .flatten()
+                .collect();
+            error.join(err);
             self.declarations.pop();
         }
 
         for output in &spec.outputs {
             if output.params.is_empty() {
-                self.add_decl_for(Declaration::Out(output.clone()));
-            } else {
-                self.add_decl_for(Declaration::ParamOut(output.clone()))
+                if let Err(e) = self.add_decl_for(Declaration::Out(output.clone())) {
+                    error.join(e);
+                }
+            } else if let Err(e) = self.add_decl_for(Declaration::ParamOut(output.clone())) {
+                error.join(e);
             }
             // Check annotated type if existing
             if let Some(output_ty) = output.annotated_type.as_ref() {
-                self.check_type(output_ty);
+                if let Err(e) = self.check_type(output_ty) {
+                    error.join(e);
+                }
             }
         }
 
-        self.check_outputs(&spec);
-        self.check_triggers(&spec);
+        if let Err(e) = self.check_outputs(spec) {
+            error.join(e);
+        }
+        if let Err(e) = self.check_triggers(spec) {
+            error.join(e);
+        }
 
-        self.result.clone()
+        Result::from(error)?;
+        Ok(self.result.clone())
     }
 
     /// Checks that if the trigger has a name, it is unique
-    fn check_triggers(&mut self, spec: &RtLolaAst) {
+    fn check_triggers(&mut self, spec: &RtLolaAst) -> Result<(), RtLolaError> {
+        let mut error = RtLolaError::new();
         for trigger in &spec.trigger {
             //Check that each supplied info stream exists
             for info_stream in &trigger.info_streams {
@@ -232,74 +279,106 @@ impl<'b> NamingAnalysis<'b> {
                     .get_decl_for(&DeclName::Ident(info_stream.name.clone()))
                 {
                     if !matches!(decl, Declaration::Out(_) | Declaration::In(_)) {
-                        self.handler.error_with_span(
-                            "Only input and output names are supported in trigger messages.",
-                            info_stream.span.clone(),
-                            Some("Found other name here"),
+                        error.add(
+                            Diagnostic::error("Only input and output names are supported in trigger messages.")
+                                .add_span_with_label(info_stream.span.clone(), Some("Found other name here"), true),
                         );
                     }
                 } else {
-                    self.handler.error_with_span(
-                        &format!("name `{}` does not exist in current scope", &info_stream.name),
-                        info_stream.span.clone(),
-                        Some("does not exist"),
+                    error.add(
+                        Diagnostic::error(&format!("name `{}` does not exist in current scope", &info_stream.name))
+                            .add_span_with_label(info_stream.span.clone(), Some("does not exist"), true),
                     );
                 }
             }
+            if let Some(pt) = trigger.annotated_pacing_type.as_ref() {
+                if let Err(e) = self.check_expression(pt) {
+                    error.join(e);
+                }
+            }
             self.declarations.push();
-            self.check_expression(&trigger.expression);
+            if let Err(e) = self.check_expression(&trigger.expression) {
+                error.join(e);
+            }
             self.declarations.pop();
         }
+        Result::from(error)
     }
 
-    fn check_outputs(&mut self, spec: &RtLolaAst) {
+    fn check_outputs(&mut self, spec: &RtLolaAst) -> Result<(), RtLolaError> {
         // recurse into expressions and check them
+        let mut error = RtLolaError::new();
         for output in &spec.outputs {
             self.declarations.push();
-            output.params.iter().for_each(|param| self.check_param(&param));
+            let para_errors = output
+                .params
+                .iter()
+                .flat_map(|param| self.check_param(param).err())
+                .flatten()
+                .collect::<RtLolaError>();
+            error.join(para_errors);
             if let Some(spawn) = &output.spawn {
                 if let Some(target) = &spawn.target {
-                    self.check_expression(target);
+                    if let Err(e) = self.check_expression(target) {
+                        error.join(e);
+                    }
                 }
                 if let Some(pacing) = &spawn.annotated_pacing {
-                    self.check_expression(&pacing);
+                    if let Err(e) = self.check_expression(pacing) {
+                        error.join(e);
+                    }
                 }
                 if let Some(cond) = &spawn.condition {
-                    self.check_expression(&cond);
+                    if let Err(e) = self.check_expression(cond) {
+                        error.join(e);
+                    }
                 }
             }
             if let Some(filter) = &output.filter {
-                self.check_expression(&filter.target);
+                if let Err(e) = self.check_expression(&filter.target) {
+                    error.join(e);
+                }
             }
             if let Some(close) = &output.close {
-                self.check_expression(&close.target);
+                if let Err(e) = self.check_expression(&close.target) {
+                    error.join(e);
+                }
+                if let Some(pacing) = &close.annotated_pacing {
+                    if let Err(e) = self.check_expression(pacing) {
+                        error.join(e);
+                    }
+                }
             }
             if let Some(pt) = output.annotated_pacing_type.as_ref() {
-                self.check_expression(&pt);
+                if let Err(e) = self.check_expression(pt) {
+                    error.join(e);
+                }
             }
             self.declarations.add_decl_for("self", Declaration::Out(output.clone()));
-            self.check_expression(&output.expression);
+            if let Err(e) = self.check_expression(&output.expression) {
+                error.join(e);
+            }
             self.declarations.pop();
         }
+        error.into()
     }
 
     /// Checks that each used identifiers has a declaration in the current or higher scope.
-    fn check_ident(&mut self, expression: &Expression, ident: &Ident) {
+    fn check_ident(&mut self, expression: &Expression, ident: &Ident) -> Result<(), Diagnostic> {
         if let Some(decl) = self.declarations.get_decl_for(&DeclName::Ident(ident.name.clone())) {
             assert!(!decl.is_type());
-
             self.result.insert(expression.id, decl);
+            Ok(())
         } else {
-            self.handler.error_with_span(
-                &format!("name `{}` does not exist in current scope", &ident.name),
-                ident.span.clone(),
-                Some("does not exist"),
-            );
+            Err(
+                Diagnostic::error(&format!("name `{}` does not exist in current scope", &ident.name))
+                    .add_span_with_label(ident.span.clone(), Some("does not exist"), true),
+            )
         }
     }
 
     /// Checks that each used function identifier has a declaration in the current scope or higher scope.
-    fn check_function(&mut self, expression: &Expression, name: &FunctionName) {
+    fn check_function(&mut self, expression: &Expression, name: &FunctionName) -> Result<(), Diagnostic> {
         let str_repr = name.to_string();
         if let Some(decl) = self.fun_declarations.get_decl_for(&DeclName::Func(name.clone().into())) {
             assert!(decl.is_function());
@@ -311,57 +390,76 @@ impl<'b> NamingAnalysis<'b> {
             // parametric outputs are represented as functions
             self.result.insert(expression.id, Declaration::ParamOut(out));
         } else {
-            self.handler.error_with_span(
-                &format!("function name `{}` does not exist in current scope", str_repr),
-                name.name.span.clone(),
-                Some("does not exist"),
+            return Err(
+                Diagnostic::error(&format!("function name `{}` does not exist in current scope", str_repr))
+                    .add_span_with_label(name.name.span.clone(), Some("does not exist"), true),
             );
         }
+        Ok(())
     }
 
-    fn check_expression(&mut self, expression: &Expression) {
+    fn check_expression(&mut self, expression: &Expression) -> Result<(), RtLolaError> {
         use self::ExpressionKind::*;
 
         match &expression.kind {
-            Ident(ident) => {
-                self.check_ident(expression, ident);
-            },
+            Ident(ident) => self.check_ident(expression, ident).map_err(RtLolaError::from),
             StreamAccess(expr, _) => self.check_expression(expr),
-            Offset(expr, _) => {
-                self.check_expression(expr);
-            },
+            Offset(expr, _) => self.check_expression(expr),
             DiscreteWindowAggregation { expr, duration, .. } => {
-                self.check_expression(expr);
-                self.check_expression(duration);
+                RtLolaError::combine(self.check_expression(expr), self.check_expression(duration), |_, _| {})
             },
             SlidingWindowAggregation { expr, duration, .. } => {
-                self.check_expression(expr);
-                self.check_expression(duration);
+                RtLolaError::combine(self.check_expression(expr), self.check_expression(duration), |_, _| {})
             },
             Binary(_, left, right) => {
-                self.check_expression(left);
-                self.check_expression(right);
+                RtLolaError::combine(self.check_expression(left), self.check_expression(right), |_, _| {})
             },
-            Lit(_) | MissingExpression => {},
+            Lit(_) | MissingExpression => Ok(()),
             Ite(condition, if_case, else_case) => {
-                self.check_expression(condition);
-                self.check_expression(if_case);
-                self.check_expression(else_case);
-            },
-            ParenthesizedExpression(_, expr, _) | Unary(_, expr) | Field(expr, _) => {
-                self.check_expression(expr);
+                let cond_errs: RtLolaError = self.check_expression(condition).into();
+                let cons_errs: RtLolaError = self.check_expression(if_case).into();
+                let alt_errs: RtLolaError = self.check_expression(else_case).into();
+                cond_errs
+                    .into_iter()
+                    .chain(cons_errs.into_iter())
+                    .chain(alt_errs.into_iter())
+                    .collect::<RtLolaError>()
+                    .into()
             },
+            ParenthesizedExpression(_, expr, _) | Unary(_, expr) | Field(expr, _) => self.check_expression(expr),
             Tuple(exprs) => {
-                exprs.iter().for_each(|expr| self.check_expression(expr));
+                exprs
+                    .iter()
+                    .flat_map(|expr| self.check_expression(expr).err())
+                    .flatten()
+                    .collect::<RtLolaError>()
+                    .into()
             },
             Function(name, types, exprs) => {
-                self.check_function(expression, name);
-                types.iter().for_each(|ty| self.check_type(ty));
-                exprs.iter().for_each(|expr| self.check_expression(expr));
+                let func_err: RtLolaError = self.check_function(expression, name).map_err(RtLolaError::from).into();
+                let type_errs: RtLolaError = types
+                    .iter()
+                    .flat_map(|ty| self.check_type(ty).err())
+                    .flatten()
+                    .collect();
+                let expr_errs: RtLolaError = exprs
+                    .iter()
+                    .flat_map(|expr| self.check_expression(expr).err())
+                    .flatten()
+                    .collect();
+                func_err
+                    .into_iter()
+                    .chain(type_errs.into_iter())
+                    .chain(expr_errs.into_iter())
+                    .collect::<RtLolaError>()
+                    .into()
             },
             Default(accessed, default) => {
-                self.check_expression(accessed);
-                self.check_expression(default);
+                RtLolaError::combine(
+                    self.check_expression(accessed),
+                    self.check_expression(default),
+                    |_, _| {},
+                )
             },
             Method(expr, name, types, args) => {
                 // Method is equal to function with `expr` as first argument
@@ -369,10 +467,28 @@ impl<'b> NamingAnalysis<'b> {
                     name: name.name.clone(),
                     arg_names: vec![None].into_iter().chain(name.arg_names.clone()).collect(),
                 };
-                self.check_function(expression, &func_name);
-                types.iter().for_each(|ty| self.check_type(ty));
-                self.check_expression(expr);
-                args.iter().for_each(|expr| self.check_expression(expr));
+                let func_errs: RtLolaError = self
+                    .check_function(expression, &func_name)
+                    .map_err(RtLolaError::from)
+                    .into();
+                let type_errs: RtLolaError = types
+                    .iter()
+                    .flat_map(|ty| self.check_type(ty).err())
+                    .flatten()
+                    .collect();
+                let expr_errs: RtLolaError = self.check_expression(expr).into();
+                let arg_errs: RtLolaError = args
+                    .iter()
+                    .flat_map(|expr| self.check_expression(expr).err())
+                    .flatten()
+                    .collect();
+                func_errs
+                    .into_iter()
+                    .chain(type_errs.into_iter())
+                    .chain(expr_errs.into_iter())
+                    .chain(arg_errs.into_iter())
+                    .collect::<RtLolaError>()
+                    .into()
             },
         }
     }
@@ -519,20 +635,18 @@ impl From<FunctionName> for crate::hir::FunctionName {
 #[cfg(test)]
 mod tests {
 
-    use std::path::PathBuf;
-
-    use rtlola_parser::{parse_with_handler, ParserConfig};
+    use rtlola_parser::{parse, ParserConfig};
 
     use super::*;
 
     /// Parses the content, runs naming analysis, and returns number of errors
     fn number_of_naming_errors(content: &str) -> usize {
-        let handler = Handler::new(PathBuf::new(), content.into());
-        let ast = parse_with_handler(ParserConfig::for_string(content.to_string()), &handler)
-            .unwrap_or_else(|e| panic!("{}", e));
-        let mut naming_analyzer = NamingAnalysis::new(&handler);
-        naming_analyzer.check(&ast);
-        handler.emitted_errors()
+        let ast = parse(ParserConfig::for_string(content.to_string())).unwrap_or_else(|e| panic!("{:?}", e));
+        let mut naming_analyzer = NamingAnalysis::new();
+        match naming_analyzer.check(&ast) {
+            Ok(_) => 0,
+            Err(e) => e.num_errors(),
+        }
     }
 
     #[test]
