@@ -2,7 +2,8 @@ use std::fmt::Debug;
 
 use itertools::Itertools;
 use num::{CheckedDiv, Integer};
-use rtlola_reporting::{Diagnostic, Handler, Span};
+use rtlola_reporting::{Diagnostic, Span};
+use rusttyc::types::ContextSensitiveVariant;
 use rusttyc::{Arity, Constructable, Partial, TcErr, TcKey, Variant};
 use uom::lib::collections::HashMap;
 use uom::lib::fmt::Formatter;
@@ -11,11 +12,11 @@ use uom::si::frequency::hertz;
 use uom::si::rational64::Frequency as UOM_Frequency;
 
 use crate::hir::{
-    AnnotatedPacingType, Constant, ExprId, Expression, ExpressionKind, Hir, Inlined, Literal, StreamAccessKind,
-    StreamReference, ValueEq,
+    AnnotatedPacingType, Constant, ExprId, Expression, ExpressionContext, ExpressionKind, Hir, Inlined, Literal,
+    StreamAccessKind, StreamReference, ValueEq,
 };
 use crate::modes::HirMode;
-use crate::type_check::rtltc::{Emittable, TypeError};
+use crate::type_check::rtltc::{Resolvable, TypeError};
 use crate::type_check::ConcretePacingType;
 
 /// The activation condition describes when an event-based stream produces a new value.
@@ -59,19 +60,6 @@ pub(crate) enum AbstractExpressionType {
     Expression(Expression),
 }
 
-impl PartialEq for AbstractExpressionType {
-    fn eq(&self, other: &Self) -> bool {
-        match (self, other) {
-            (AbstractExpressionType::Any, AbstractExpressionType::Any) => true,
-            (AbstractExpressionType::AnyClose, AbstractExpressionType::AnyClose) => true,
-            (AbstractExpressionType::Expression(l), AbstractExpressionType::Expression(r)) => l.value_eq(r),
-            _ => false,
-        }
-    }
-}
-
-impl Eq for AbstractExpressionType {}
-
 /// The internal representation of the overall Stream pacing
 /// Types are given by keys in the respective rustic instances
 #[derive(Clone, Copy, Debug, PartialEq, Eq)]
@@ -90,7 +78,8 @@ pub(crate) struct StreamTypeKeys {
 /// Reference for stream  template during pacing type inference, used in error reporting.
 #[derive(Debug)]
 pub(crate) struct InferredTemplates {
-    pub(crate) spawn: Option<(ConcretePacingType, Expression)>,
+    pub(crate) spawn_pacing: Option<ConcretePacingType>,
+    pub(crate) spawn_cond: Option<Expression>,
     pub(crate) filter: Option<Expression>,
     pub(crate) close: Option<Expression>,
 }
@@ -114,6 +103,19 @@ pub(crate) enum PacingErrorKind {
     UnintuitivePacingWarning(Span, ConcretePacingType),
     Other(Span, String, Vec<Box<dyn PrintableVariant>>),
     SpawnPeriodicMismatch(Span, Span, (ConcretePacingType, Expression)),
+    InvalidSyncAccessParameter {
+        target_span: Span,
+        target_spawn_expr: Expression,
+        own_spawn_expr: Expression,
+        arg: Expression,
+    },
+    NonParamInSyncAccess(Span),
+    ParameterAmountMismatch {
+        target_span: Span,
+        exp_span: Span,
+        given_num: usize,
+        expected_num: usize,
+    },
 }
 
 impl std::ops::BitAnd for ActivationCondition {
@@ -275,7 +277,7 @@ impl ActivationCondition {
         use ActivationCondition::*;
         match self {
             True => "⊤".into(),
-            Stream(sr) => stream_names[&sr].into(),
+            Stream(sr) => stream_names[sr].into(),
             Conjunction(childs) => {
                 let child_string: String = childs.iter().map(|ac| ac.to_string(stream_names)).join(" ∧ ");
                 format!("({})", child_string)
@@ -288,41 +290,36 @@ impl ActivationCondition {
     }
 }
 
-impl Emittable for PacingErrorKind {
-    fn emit(
+impl Resolvable for PacingErrorKind {
+    fn into_diagnostic(
         self,
-        handler: &Handler,
         spans: &[&HashMap<TcKey, Span>],
         names: &HashMap<StreamReference, &str>,
         key1: Option<TcKey>,
         key2: Option<TcKey>,
-    ) {
+    ) -> Diagnostic {
         let pacing_spans = spans[0];
         let exp_spans = spans[1];
         use PacingErrorKind::*;
         match self {
             FreqAnnotationNeeded(span) => {
-                handler.error_with_span(
-                    "In pacing type analysis:\nFrequency annotation needed.",
+                Diagnostic::error("In pacing type analysis:\nFrequency annotation needed.").add_span_with_label(
                     span,
                     Some("here"),
-                );
+                    true,
+                )
             },
             NeverEval(span) => {
-                Diagnostic::error(
-                    handler,
-                    "In pacing type analysis:\nThe following stream is never evaluated.",
-                )
-                .add_span_with_label(span, Some("here"), true)
-                .add_note("Help: Consider annotating a pacing type explicitly.")
-                .emit();
+                Diagnostic::error("In pacing type analysis:\nThe following stream or expression is never evaluated.")
+                    .add_span_with_label(span, Some("here"), true)
+                    .add_note("Help: Consider annotating a pacing type explicitly.")
             },
             MalformedAc(span, reason) => {
-                handler.error_with_span(
-                    &format!("In pacing type analysis:\nMalformed activation condition: {}", reason),
-                    span,
-                    Some("here"),
-                );
+                Diagnostic::error(&format!(
+                    "In pacing type analysis:\nMalformed activation condition: {}",
+                    reason
+                ))
+                .add_span_with_label(span, Some("here"), true)
             },
             MixedEventPeriodic(absty1, absty2) => {
                 let span1 = key1.and_then(|k| pacing_spans.get(&k).cloned());
@@ -330,7 +327,6 @@ impl Emittable for PacingErrorKind {
                 let ty1 = absty1.to_pretty_string(names);
                 let ty2 = absty2.to_pretty_string(names);
                 Diagnostic::error(
-                    handler,
                     format!(
                         "In pacing type analysis:\nMixed an event and a periodic type: {} and {}",
                         ty1, ty2
@@ -338,14 +334,16 @@ impl Emittable for PacingErrorKind {
                     .as_str(),
                 )
                 .maybe_add_span_with_label(span1, Some(format!("Found {} here", ty1).as_str()), true)
-                .maybe_add_span_with_label(span2, Some(format!("and found {} here", ty2).as_str()), false)
-                .emit();
+                .maybe_add_span_with_label(
+                    span2,
+                    Some(format!("and found {} here", ty2).as_str()),
+                    false,
+                )
             },
             IncompatibleExpressions(e1, e2) => {
                 let span1 = key1.and_then(|k| exp_spans.get(&k).cloned());
                 let span2 = key2.and_then(|k| exp_spans.get(&k).cloned());
                 Diagnostic::error(
-                    handler,
                     format!(
                         "In pacing type analysis:\nIncompatible expressions: {} and {}",
                         e1.to_pretty_string(names),
@@ -363,11 +361,9 @@ impl Emittable for PacingErrorKind {
                     Some(format!("and found {} here", e2.to_pretty_string(names)).as_str()),
                     false,
                 )
-                .emit();
             },
             UnintuitivePacingWarning(span, inferred) => {
                 Diagnostic::warning(
-                    handler,
                     format!(
                         "In pacing type analysis:\nInferred complex pacing type: {}",
                         inferred.to_pretty_string(names)
@@ -382,11 +378,9 @@ impl Emittable for PacingErrorKind {
                     )
                     .as_str(),
                 )
-                .emit()
             },
             Other(span, reason, causes) => {
                 Diagnostic::error(
-                    handler,
                     format!(
                         "In pacing type analysis:\n{} {}",
                         reason,
@@ -396,39 +390,49 @@ impl Emittable for PacingErrorKind {
                 )
                 .add_span_with_label(span, Some("here"), true)
                 .maybe_add_span_with_label(key1.and_then(|k| pacing_spans.get(&k).cloned()), Some("here"), true)
-                .maybe_add_span_with_label(key2.and_then(|k| pacing_spans.get(&k).cloned()), Some("here"), true)
-                .emit();
+                .maybe_add_span_with_label(
+                    key2.and_then(|k| pacing_spans.get(&k).cloned()),
+                    Some("here"),
+                    true,
+                )
             },
             ParameterizationNotAllowed(span) => {
                 Diagnostic::error(
-                    handler,
                     "In pacing type analysis:\nSynchronous access to a parameterized stream is not allowed here.",
                 )
                 .add_span_with_label(span, Some("here"), true)
                 .add_note("Help: Consider using a hold access")
-                .emit();
             },
             ParameterizationNeeded { who, why, inferred } => {
-                let InferredTemplates { spawn, filter, close } = *inferred;
-                let spawn_str = spawn.map_or("".into(), |(pacing, cond)| {
-                    format!(
-                        "\nspawn @{} with <...> if {}",
-                        pacing.to_pretty_string(names),
-                        cond.pretty_string(names)
-                    )
-                });
+                let InferredTemplates {
+                    spawn_pacing,
+                    spawn_cond,
+                    filter,
+                    close,
+                } = *inferred;
+                let spawn_str = match (spawn_pacing, spawn_cond) {
+                    (Some(pacing), Some(cond)) => {
+                        format!(
+                            "\nspawn @{} with <...> if {}",
+                            pacing.to_pretty_string(names),
+                            cond.pretty_string(names)
+                        )
+                    },
+                    (Some(pacing), None) => format!("\nspawn @{} with <...>", pacing.to_pretty_string(names)),
+                    (None, Some(cond)) => format!("\nspawn <...> if {}", cond.pretty_string(names)),
+                    (None, None) => "".to_string(),
+                };
                 let filter_str: String =
                     filter.map_or("".into(), |filter| format!("\nfilter {}", filter.pretty_string(names)));
                 let close_str: String =
                     close.map_or("".into(), |close| format!("\nclose {}", close.pretty_string(names)));
-                Diagnostic::error(handler, "In pacing type analysis:\nParameterization needed")
+                Diagnostic::error("In pacing type analysis:\nParameterization needed")
                     .add_span_with_label(who, Some("here"), true)
                     .add_span_with_label(why, Some("As of synchronous access occurring here"), false)
                     .add_note(&format!(
                         "Help: Consider adding the following template annotations:{}{}{}",
                         spawn_str, filter_str, close_str,
                     ))
-                    .emit();
             },
             PacingTypeMismatch(bound, inferred) => {
                 let bound_str = bound.to_pretty_string(names);
@@ -436,7 +440,6 @@ impl Emittable for PacingErrorKind {
                 let bound_span = key1.map(|k| pacing_spans[&k].clone());
                 let inferred_span = key2.and_then(|k| pacing_spans.get(&k).cloned());
                 Diagnostic::error(
-                    handler,
                     format!(
                         "In pacing type analysis:\nInferred pacing type: {} but expected: {}",
                         &inferred_str, &bound_str
@@ -449,10 +452,8 @@ impl Emittable for PacingErrorKind {
                     Some(format!("Inferred {} here", inferred_str).as_str()),
                     true,
                 )
-                .emit();
             },
             SpawnPeriodicMismatch(access_span, target_span, (access_pacing, access_condition)) => Diagnostic::error(
-                handler,
                 "In pacing type analysis:\nPeriodic stream out of sync with accessed stream due to a spawn annotation.",
             )
             .add_span_with_label(
@@ -467,8 +468,40 @@ impl Emittable for PacingErrorKind {
                 ),
                 true,
             )
-            .add_span_with_label(target_span, Some("Found target stream here"), false)
-            .emit(),
+            .add_span_with_label(target_span, Some("Found target stream here"), false),
+            InvalidSyncAccessParameter {
+                target_span,
+                target_spawn_expr,
+                own_spawn_expr,
+                arg,
+            } => {
+                let target_expr = target_spawn_expr.pretty_string(names);
+                let own_expr = own_spawn_expr.pretty_string(names);
+                let supplied = arg.pretty_string(names);
+
+                Diagnostic::error(
+                    "In pacing type analysis:\nInvalid argument for synchronized access:"
+                )
+                .add_span_with_label(target_span, Some(&format!("Target expected the argument to be equal to the spawn expression: ({})", target_expr)), false)
+                .add_span_with_label(arg.span, Some(&format!("Supplied arguments ({}) resolved to the spawn expressions: ({})",
+                                                             supplied,
+                    own_expr
+                )), true)
+                    .add_note("Note: Each parameter of the accessed stream requires a counterpart which is a parameter of the accessing stream.")
+            },
+            NonParamInSyncAccess(span) => {
+                Diagnostic::error(
+                                  "In pacing type analysis:\nOnly parameters are allowed as arguments when synchronously accessing a stream:"
+                )
+                    .add_span_with_label(span, Some("Found an expression that is not a parameter here"), true)
+            },
+            ParameterAmountMismatch { target_span, exp_span, given_num, expected_num} => {
+                Diagnostic::error(
+                                  "In pacing type analysis:\nMismatch between number of given arguments and expected spawn arguments:"
+                )
+                    .add_span_with_label(exp_span, Some(&format!("Got {} arguments here.", given_num)), true)
+                    .add_span_with_label(target_span, Some(&format!("Expected {} arguments here.", expected_num)), false)
+            }
         }
     }
 }
@@ -477,7 +510,9 @@ pub(crate) trait PrintableVariant: Debug {
     fn to_pretty_string(&self, names: &HashMap<StreamReference, &str>) -> String;
 }
 
-impl<V: 'static + Variant<Err = PacingErrorKind> + PrintableVariant> From<TcErr<V>> for TypeError<PacingErrorKind> {
+impl<V: 'static + ContextSensitiveVariant<Err = PacingErrorKind> + PrintableVariant> From<TcErr<V>>
+    for TypeError<PacingErrorKind>
+{
     fn from(err: TcErr<V>) -> TypeError<PacingErrorKind> {
         match err {
             TcErr::KeyEquation(k1, k2, err) => {
@@ -532,6 +567,9 @@ impl<V: 'static + Variant<Err = PacingErrorKind> + PrintableVariant> From<TcErr<
                     key2: preliminary.children[idx],
                 }
             },
+            TcErr::CyclicGraph => {
+                panic!("Cyclic pacing type constraint system");
+            },
         }
     }
 }
@@ -546,8 +584,7 @@ impl std::fmt::Display for Freq {
                 write!(
                     f,
                     "{}",
-                    freq.clone()
-                        .into_format_args(hertz, uom::fmt::DisplayStyle::Abbreviation)
+                    (*freq).into_format_args(hertz, uom::fmt::DisplayStyle::Abbreviation)
                 )
             },
         }
@@ -694,21 +731,22 @@ impl AbstractPacingType {
     }
 }
 
-impl Variant for AbstractExpressionType {
+impl ContextSensitiveVariant for AbstractExpressionType {
+    type Context = ExpressionContext;
     type Err = PacingErrorKind;
 
     fn top() -> Self {
         AbstractExpressionType::Any
     }
 
-    fn meet(lhs: Partial<Self>, rhs: Partial<Self>) -> Result<Partial<Self>, Self::Err> {
+    fn meet(lhs: Partial<Self>, rhs: Partial<Self>, ctx: &ExpressionContext) -> Result<Partial<Self>, Self::Err> {
         assert_eq!(lhs.least_arity, 0, "suspicious child");
         assert_eq!(rhs.least_arity, 0, "suspicious child");
         let new_var = match (lhs.variant, rhs.variant) {
             (Self::Any, x) | (x, Self::Any) => Ok(x),
             (Self::AnyClose, x) | (x, Self::AnyClose) => Ok(x),
             (Self::Expression(a), Self::Expression(b)) => {
-                if a.value_eq(&b) {
+                if a.value_eq(&b, ctx) {
                     Ok(Self::Expression(a))
                 } else {
                     Err(PacingErrorKind::IncompatibleExpressions(
@@ -724,9 +762,18 @@ impl Variant for AbstractExpressionType {
         })
     }
 
-    fn arity(&self) -> Arity {
+    fn arity(&self, _ctx: &ExpressionContext) -> Arity {
         Arity::Fixed(0)
     }
+
+    fn equal(this: &Self, that: &Self, ctx: &Self::Context) -> bool {
+        match (this, that) {
+            (AbstractExpressionType::Expression(e1), AbstractExpressionType::Expression(e2)) => e1.value_eq(e2, ctx),
+            (AbstractExpressionType::Any, AbstractExpressionType::Any)
+            | (AbstractExpressionType::AnyClose, AbstractExpressionType::AnyClose) => true,
+            _ => false,
+        }
+    }
 }
 
 impl Constructable for AbstractExpressionType {
@@ -771,7 +818,7 @@ impl ConcretePacingType {
         match self {
             ConcretePacingType::Event(ac) => ac.to_string(names),
             ConcretePacingType::FixedPeriodic(freq) => {
-                freq.clone()
+                (*freq)
                     .into_format_args(hertz, uom::fmt::DisplayStyle::Abbreviation)
                     .to_string()
             },
